#ifndef CORE_LOGGER_HPP
#define CORE_LOGGER_HPP

#include <cstdio>

#include "core/Terminal.hpp"
#include "core/ToString.hpp"

enum class LogLevel { NONE, ERROR, WARNING, INFO, DEBUG };

extern LogLevel logLevel;

using ReportHandler = void (*)(
    LogLevel l, const char* file, int line, void* data, const char* message);
void callReportHandler(
    LogLevel l, const char* file, int line, const char* report);
void setReportHandler(ReportHandler h, void* data);

template<typename... Args>
void callReportHandler(
    LogLevel l, const char* file, int line, const char* format,
    Args&&... args) {
    char buffer[512];
    formatBuffer(buffer, sizeof(buffer), format, Core::forward<Args>(args)...);
    callReportHandler(l, file, line, buffer);
}

#define REPORT(l, ...) callReportHandler(l, __FILE__, __LINE__, __VA_ARGS__)

const char* getShortFileName(const char* s);

template<typename... Args>
void printLog(
    LogLevel l, const char* file, int line, const char* prefix,
    const char* format, Args&&... args) {
    if(logLevel < l) {
        return;
    }
    file = getShortFileName(file);
    fputs(prefix, stdout);
    printf("%s:%d | ", file, line);
    char buffer[512];
    formatBuffer(buffer, sizeof(buffer), format, Core::forward<Args>(args)...);
    fputs(buffer, stdout);
    puts(TERMINAL_RESET);
}

#if defined(LOG_LEVEL) && LOG_LEVEL >= 1
#define LOG_ERROR(...)                                                       \
    printLog(                                                                \
        LogLevel::ERROR, __FILE__, __LINE__, TERMINAL_BRIGHT_RED "[ERROR] ", \
        __VA_ARGS__)
#else
#define LOG_ERROR(...)
#endif

#if defined(LOG_LEVEL) && LOG_LEVEL >= 2
#define LOG_WARNING(...)                                  \
    printLog(                                             \
        LogLevel::WARNING, __FILE__, __LINE__,            \
        TERMINAL_BRIGHT_YELLOW "[WARNING] ", __VA_ARGS__)
#else
#define LOG_WARNING(...)
#endif

#if defined(LOG_LEVEL) && LOG_LEVEL >= 3
#define LOG_INFO(...)                                                \
    printLog(                                                        \
        LogLevel::INFO, __FILE__, __LINE__, TERMINAL_BOLD "[INFO] ", \
        __VA_ARGS__)
#else
#define LOG_INFO(...)
#endif

#if defined(LOG_LEVEL) && LOG_LEVEL >= 4
#define LOG_DEBUG(...)                                               \
    printLog(                                                        \
        LogLevel::DEBUG, __FILE__, __LINE__,                         \
        TERMINAL_BOLD TERMINAL_BRIGHT_BLACK "[DEBUG] ", __VA_ARGS__)
#else
#define LOG_DEBUG(...)
#endif

#endif
