#ifndef CORE_PROBING_HASHMAP_HPP
#define CORE_PROBING_HASHMAP_HPP

#include "core/List.hpp"
#include "core/Logger.hpp"
#include "core/ToString.hpp"
#include "core/Utility.hpp"

template<typename T>
concept Hashable = requires(const T& t) { t.hashCode(); };
template<typename T>
concept HashCast = requires(const T& t) { static_cast<size_t>(t); };

namespace Core {
    template<Hashable H>
    inline size_t hashCode(const H& key) {
        return key.hashCode();
    }

    template<HashCast H>
    inline size_t hashCode(const H& key) {
        return static_cast<size_t>(key);
    }

    template<typename K, typename V>
    struct ProbingHashMap final {
        template<typename Value>
        class Node final {
            friend ProbingHashMap;
            friend List<Node>;
            K key;

        public:
            Value& value;

            const K& getKey() const {
                return key;
            }

            size_t toString(char* s, size_t n) const {
                size_t total = 0;
                addString(key, s, n, total);
                addString(" = ", s, n, total);
                addString(value, s, n, total);
                return total;
            }

        private:
            Node(const K& key_, Value& value_) : key(key_), value(value_) {
            }
        };

    private:
        static constexpr K INVALID = {};

        template<typename Value, typename R, R (*A)(const K&, Value&)>
        class Iterator final {
            const K* currentKey;
            const K* endKey;
            Value* currentValue;

        public:
            Iterator(
                const K* key, const K* endKey_, Value* value, bool invalidSet) :
                currentKey(key), endKey(endKey_), currentValue(value) {
                if(!invalidSet) {
                    skip();
                }
            }

            Iterator& operator++() {
                ++currentKey;
                ++currentValue;
                skip();
                return *this;
            }

            bool operator!=(const Iterator& other) const {
                return currentKey != other.currentKey;
            }

            R operator*() const {
                return A(*currentKey, *currentValue);
            }

        private:
            void skip() {
                while(currentKey != endKey && *currentKey == INVALID) {
                    ++currentKey;
                    ++currentValue;
                }
            }
        };

        template<typename Value>
        static Node<Value> access(const K& key, Value& value) {
            return Node<Value>(key, value);
        }

        template<typename Value>
        static Value& accessValue(const K&, Value& value) {
            return value;
        }

        static const K& accessKey(const K& key, const V&) {
            return key;
        }

        template<typename Value>
        using BaseEntryIterator = Iterator<Value, Node<Value>, access<Value>>;
        using EntryIterator = BaseEntryIterator<V>;
        using ConstEntryIterator = BaseEntryIterator<const V>;

        template<typename Value>
        using BaseValueIterator = Iterator<Value, Value&, accessValue<Value>>;
        using ValueIterator = BaseValueIterator<V>;
        using ConstValueIterator = BaseValueIterator<const V>;

        using ConstKeyIterator = Iterator<const V, const K&, accessKey>;

        template<typename M, typename I>
        struct IteratorAdapter final {
            M& map;

            I begin() const {
                return {
                    map.keys.begin(), map.keys.end(), map.values,
                    map.invalidSet};
            }

            I end() const {
                return {
                    map.keys.end(), map.keys.end(), nullptr, map.invalidSet};
            }
        };

        using ValueIteratorAdapter =
            IteratorAdapter<ProbingHashMap, ValueIterator>;
        using ConstValueIteratorAdapter =
            IteratorAdapter<const ProbingHashMap, ConstValueIterator>;

        using ConstKeyIteratorAdapter =
            IteratorAdapter<const ProbingHashMap, ConstKeyIterator>;

    private:
        List<K> keys{};
        V* values = nullptr;
        List<int> jumps{};
        size_t entries = 0;
        bool invalidSet = false;

    public:
        ProbingHashMap() = default;

        ProbingHashMap(const ProbingHashMap& other) {
            for(const auto& e : other) {
                add(e.getKey(), e.value);
            }
        }

        ProbingHashMap(ProbingHashMap&& other) {
            swap(other);
        }

        ~ProbingHashMap() {
            size_t length = keys.getLength();
            if(length > 0) {
                for(size_t i = 1; i < length; i++) {
                    if(keys[i] != INVALID) {
                        values[i].~V();
                    }
                }
                if(invalidSet) {
                    values[length].~V();
                }
            }
            delete[] reinterpret_cast<AlignedType<V>*>(values);
        }

        ProbingHashMap& operator=(ProbingHashMap other) {
            swap(other);
            return *this;
        }

        void rehash(size_t minCapacity) {
            if(minCapacity <= keys.getLength()) {
                return;
            }
            ProbingHashMap<K, V> map;
            size_t l = (1lu << roundUpLog2(max(minCapacity, 8lu))) + 1;
            map.keys.resize(l, INVALID);
            map.values = reinterpret_cast<V*>(coreNewN(AlignedType<V>, l));
            map.jumps.resize(l, INVALID);
            size_t length = keys.getLength();
            if(length > 0) {
                for(size_t i = 1; i < length; i++) {
                    if(keys[i] != INVALID) {
                        map.add(keys[i], values[i]);
                    }
                }
                if(invalidSet) {
                    map.add(INVALID, values[length]);
                }
            }
            swap(map);
        }

        template<typename... Args>
        bool tryEmplace(V*& v, const K& key, Args&&... args) {
            size_t index = 0;
            if(key == INVALID) {
                if(invalidSet) {
                    return false;
                }
                rehash(1);
                invalidSet = true;
            } else {
                index = searchSlot(key);
                if(keys[index] == key) {
                    return false;
                }
            }
            keys[index] = key;
            v = new(values + index) V(Core::forward<Args>(args)...);
            entries++;
            markSlot(key);
            return true;
        }

        template<typename VA>
        V& put(const K& key, VA&& value) {
            size_t index = 0;
            if(key == INVALID) {
                if(invalidSet) {
                    return (values[0] = Core::forward<VA>(value));
                }
                rehash(1);
                invalidSet = true;
            } else {
                index = searchSlot(key);
                if(keys[index] == key) {
                    return (values[index] = Core::forward<VA>(value));
                }
            }
            new(values + index) V(Core::forward<VA>(value));
            entries++;
            keys[index] = key;
            markSlot(key);
            return values[index];
        }

        template<typename VA>
        ProbingHashMap& add(const K& key, VA&& value) {
            put(key, Core::forward<VA>(value));
            return *this;
        }

        bool remove(const K& key) {
            size_t index = 0;
            if(key == INVALID) {
                if(!invalidSet) {
                    return false;
                }
                invalidSet = false;
            } else {
                index = searchSlot(key);
                if(keys[index] != key) {
                    return false;
                }
            }
            values[index].~V();
            entries--;
            demarkSlot(key);
            keys[index] = INVALID;
            return true;
        }

        const V* search(const K& key) const {
            return searchValue<const V>(key);
        }

        V* search(const K& key) {
            return searchValue<V>(key);
        }

        bool contains(const K& key) const {
            return search(key) != nullptr;
        }

        ProbingHashMap& clear() {
            ProbingHashMap<K, V> map;
            swap(map);
            return *this;
        }

        ConstKeyIteratorAdapter getKeys() const {
            return {*this};
        }

        ValueIteratorAdapter getValues() {
            return {*this};
        }

        ConstValueIteratorAdapter getValues() const {
            return {*this};
        }

        EntryIterator begin() {
            return {keys.begin(), keys.end(), values, invalidSet};
        }

        EntryIterator end() {
            return {keys.end(), keys.end(), nullptr, invalidSet};
        }

        ConstEntryIterator begin() const {
            return {keys.begin(), keys.end(), values, invalidSet};
        }

        ConstEntryIterator end() const {
            return {keys.end(), keys.end(), nullptr, invalidSet};
        }

        void swap(ProbingHashMap& o) {
            Core::swap(o.keys, keys);
            Core::swap(o.values, values);
            Core::swap(o.jumps, jumps);
            Core::swap(o.entries, entries);
            Core::swap(o.invalidSet, invalidSet);
        }

    private:
        static constexpr size_t MAX_CLUSTER = 5;

        size_t searchSlot(const K& key) {
            rehash(1);
            while(true) {
                size_t baseHash = hashCode(key) * 514'685'581u;
                size_t end = keys.getLength() - 2;
                // rehash on bad clustering
                for(size_t i = 0; i <= MAX_CLUSTER; i++) {
                    size_t hash = 1 + ((baseHash + i) & end);
                    if((keys[hash] == INVALID && jumps[hash] == 0) ||
                       keys[hash] == key) {
                        return hash;
                    }
                }
                rehash(keys.getLength() + 1);
            }
        }

        void markSlot(const K& key) {
            size_t baseHash = hashCode(key) * 514'685'581u;
            size_t end = keys.getLength() - 2;
            for(size_t i = 0; i <= MAX_CLUSTER; i++) {
                size_t hash = 1 + ((baseHash + i) & end);
                if(keys[hash] == key) {
                    return;
                }
                jumps[hash]++;
            }
        }

        void demarkSlot(const K& key) {
            size_t baseHash = hashCode(key) * 514'685'581u;
            size_t end = keys.getLength() - 2;
            for(size_t i = 0; i <= MAX_CLUSTER; i++) {
                size_t hash = 1 + ((baseHash + i) & end);
                if(keys[hash] == key) {
                    return;
                }
                jumps[hash]--;
            }
        }

        template<typename Value>
        Value* searchValue(const K& key) const {
            if(keys.getLength() != 0) {
                if(key == INVALID) {
                    return invalidSet ? values : nullptr;
                }
                size_t baseHash = hashCode(key) * 514'685'581u;
                size_t end = keys.getLength() - 2;
                for(size_t i = 0; i <= end; i++) [[unlikely]] {
                    size_t hash = 1 + ((baseHash + i) & end);
                    if(keys[hash] == key) [[likely]] {
                        return values + hash;
                    } else if(jumps[hash] == 0) {
                        return nullptr;
                    }
                }
            }
            return nullptr;
        }
    };
}

#endif
