#include "core/Frustum.hpp"

#include <cmath>

using Core::Frustum;
using V3 = Core::Vector3;
using V4 = Core::Vector4;

Frustum::Frustum(float fieldOfView, float nearClip, float farClip) :
    projection(), planes(), tan(tanf(fieldOfView * 0.5f)), near(nearClip),
    far(farClip) {
    float diff = 1.0f / (near - far);
    projection.set(1, V4(0.0f, 1.0f / tan, 0.0f, 0.0f));
    projection.set(
        2, V4(0.0f, 0.0f, (near + far) * diff, (2.0f * near * far) * diff));
    projection.set(3, V4(0.0f, 0.0f, -1.0f, 0.0f));
}

const Core::Matrix& Frustum::updateProjection(const IntVector2& size) {
    projection.set(
        0, V4(static_cast<float>(size[1]) / (tan * static_cast<float>(size[0])),
              0.0f, 0.0f, 0.0f));
    return projection;
}

void Frustum::updatePlanes(
    const V3& pos, const V3& right, const V3& up, const V3& front,
    const IntVector2& size) {
    float aspect = static_cast<float>(size[0]) / static_cast<float>(size[1]);

    float hNearHeight = tan * near;
    float hNearWidth = hNearHeight * aspect;

    float hFarHeight = tan * far;
    float hFarWidth = hFarHeight * aspect;

    V3 fCenter = pos + front * far;
    V3 upFar = up * hFarHeight;
    V3 rightFar = right * hFarWidth;
    V3 fTopLeft = fCenter + upFar - rightFar;
    V3 fTopRight = fCenter + upFar + rightFar;
    V3 fBottomRight = fCenter - upFar + rightFar;

    V3 nCenter = pos + front * near;
    V3 upNear = up * hNearHeight;
    V3 rightNear = right * hNearWidth;
    V3 nTopLeft = nCenter + upNear - rightNear;
    V3 nBottomLeft = nCenter - upNear - rightNear;
    V3 nBottomRight = nCenter - upNear + rightNear;

    planes[0] = Plane(nBottomRight, nTopLeft, nBottomLeft);     // near plane
    planes[1] = Plane(fTopRight, fBottomRight, fTopLeft);       // far plane
    planes[2] = Plane(nBottomRight, nBottomLeft, fBottomRight); // bottom plane
    planes[3] = Plane(fTopLeft, nTopLeft, fTopRight);           // top plane
    planes[4] = Plane(nBottomLeft, nTopLeft, fTopLeft);         // left plane
    planes[5] = Plane(fBottomRight, fTopRight, nBottomRight);   // right plane
}

bool Frustum::isInside(const V3& pos) const {
    for(const Plane& p : planes) {
        if(p.signedDistance(pos) < 0.0f) {
            return false;
        }
    }
    return true;
}

bool Frustum::isInside(const V3& pos, float radius) const {
    for(const Plane& p : planes) {
        if(p.signedDistance(pos) < -radius) {
            return false;
        }
    }
    return true;
}
