#include "core/Test.hpp"

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "core/Logger.hpp"
#include "core/Utility.hpp"

typedef struct {
    char* file;
    int tests;
    int successTests;
} Result;

static Result* results = nullptr;
static size_t resultsIndex = 0;
static size_t resultsCapacity = 0;

static Result* getResult(const char* file) {
    for(size_t i = 0; i < resultsIndex; i++) {
        Result* r = results + i;
        if(strcmp(r->file, file) == 0) {
            return r;
        }
    }
    while(resultsIndex >= resultsCapacity) {
        size_t newCapacity = resultsCapacity == 0 ? 8 : resultsCapacity * 2;
        results = static_cast<Result*>(
            Core::reallocateRaw(results, newCapacity * sizeof(Result)));
        resultsCapacity = newCapacity;
    }
    Result* r = results + (resultsIndex++);
    r->file = strdup(file);
    r->successTests = 0;
    r->tests = 0;
    return r;
}

void Core::finalizeTests(void) {
    for(size_t i = 0; i < resultsIndex; i++) {
        Result* r = results + i;
        bool c = r->successTests == r->tests;
        fputs(c ? TERMINAL_GREEN : TERMINAL_RED, stdout);
        printf(
            "%s - %d / %d tests succeeded", r->file, r->successTests, r->tests);
        puts(TERMINAL_RESET);
        free(r->file);
    }
    Core::deallocateRaw(results);
    results = nullptr;
    resultsIndex = 0;
    resultsCapacity = 0;
}

bool Core::addTestResult(const char* file, bool comparison) {
    Result* r = getResult(file);
    r->tests++;
    if(comparison) {
        r->successTests++;
        return true;
    }
    return false;
}

#define TEST_SUCCESS(result)                \
    file = getShortFileName(file);          \
    if(Core::addTestResult(file, result)) { \
        return true;                        \
    }

bool Core::testString(
    const char* file, int line, const char* wanted, const char* actual) {
    TEST_SUCCESS(strcmp(wanted, actual) == 0)
    fputs(TERMINAL_RED, stdout);
    printf("%s:%d - expected '%s' got '%s'", file, line, wanted, actual);
    puts(TERMINAL_RESET);
    return false;
}

bool Core::testFloat(
    const char* file, int line, float wanted, float actual, float error) {
    float diff = wanted - actual;
    diff = diff < 0.0f ? -diff : diff;
    TEST_SUCCESS(diff <= error)
    fputs(TERMINAL_RED, stdout);
    printf(
        "%s:%d - expected '%.3f' got '%.3f'", file, line,
        static_cast<double>(wanted), static_cast<double>(actual));
    puts(TERMINAL_RESET);
    return false;
}

bool Core::testNull(const char* file, int line, const void* actual) {
    TEST_SUCCESS(actual == nullptr)
    fputs(TERMINAL_RED, stdout);
    printf("%s:%d - expected null", file, line);
    puts(TERMINAL_RESET);
    return false;
}

bool Core::testNotNull(const char* file, int line, const void* actual) {
    TEST_SUCCESS(actual != nullptr)
    fputs(TERMINAL_RED, stdout);
    printf("%s:%d - expected valid pointer", file, line);
    puts(TERMINAL_RESET);
    return false;
}
